psi := function(v,q,t)
	local basis,d1;
	basis := BasisVectors(Basis(AsVectorSpace(GF(q),GF(q^t))));
	d1 := Length(v)/t;
	return List([1..d1],i->v{[(i-1)*t+1..i*t]}*basis);
end;

BilinearFormByFieldReduction:=function(bil1,f2)
	# f2 is a subfield of the basefield of the bilinear form bil1
	local mat1,f1,basis,basvecs,d1,t,d2,V2,V1,phi,b2,b2vecs,mat2,i,row,j,bil2;
	mat1:=bil1!.matrix;
	f1:=bil1!.basefield;
	basis:=Basis(AsVectorSpace(f2,f1));
	basvecs:=BasisVectors(basis);
	d1:=Size(mat1);
	t:=Dimension(AsVectorSpace(f2,f1));
	d2:=d1*t;
	V2:=f2^d2;
	V1:=f1^d1;
	phi:=function(v) # is an isomorphism from V2 to V1
		return List([1..d1],i->v{[(i-1)*t+1..i*t]}*basvecs);
	end;
	# Here is a test to see if this is a bijection:
	# AsSet(Elements(V1))=AsSet(List(Elements(V2),v->phi(v)));
	b2:=Basis(V2);
	b2vecs:=BasisVectors(b2);
	mat2:=[];
	for i in [1..d2] do 
		row:=[];
		for j in [1..d2] do
			Add(row,Trace(f1,f2,phi(b2vecs[i])*mat1*phi(b2vecs[j])));
		od;
		Add(mat2,row);
	od;
	bil2:=SesquilinearForm(PolarityOfProjectiveSpace(mat2,f2));
	return bil2;
end;

QuadraticFormByFieldReduction:=function(qf1,f2)
	# f2 is a subfield of the basefield for the quadratic form q1
	local f1,basis,basvecs,d1,t,d2,V2,V1,phi,b2,b2vecs,mat2,i,bil1,j,qf2;
	f1:=qf1!.basefield;
	basis:=Basis(AsVectorSpace(f2,f1));
	basvecs:=BasisVectors(basis);
	d1:=Size(qf1!.matrix);
	t:=Dimension(AsVectorSpace(f2,f1));
	d2:=d1*t;
	V2:=f2^d2;
	V1:=f1^d1;
	phi:=function(v) # is an isomorphism from V2 to V1
		return List([1..d1],i->v{[(i-1)*t+1..i*t]}*basvecs);
	end;
	# Here is a test to see if this is a bijection:
	# AsSet(Elements(V1))=AsSet(List(Elements(V2),v->phi(v)));
	b2:=Basis(V2);
	b2vecs:=BasisVectors(b2);
	mat2:=IdentityMat(d2,f2);
	for i in [1..d2] do
		mat2[i][i]:=Trace(f1,f2,(phi(b2vecs[i]))^qf1);
	od;
	for i in [1..d2-1] do
		for j in [i+1..d2] do
			mat2[i][j]:=Trace(f1,f2,(phi(b2vecs[i]+b2vecs[j]))^qf1
							-(phi(b2vecs[i]))^qf1-(phi(b2vecs[j]))^qf1);
			#mat2[j][i]:=mat2[i][j]; THESE entries need to be zero
		od;
	od;
	qf2:=QuadraticFormByMatrix(mat2,f2);
	return qf2;
end;

# test 1
list:=[];
for q in [2,3,4,5,8] do for t in [2,3,4] do for n in [1,2,3,4] do
f1:=GF(q^t);
f2:=GF(q);
#ps1:=HyperbolicQuadric(2*n-1,f1);
ps1:=EllipticQuadric(2*n-1,f1);
qf1:=QuadraticForm(ps1);
qf2:=QuadraticFormByFieldReduction(qf1,f2);
Add(list,PolarSpaceType(PolarSpace(qf2)));
od;od;od;
list; # contains only one type: the type of ps1

# test 2, using SesquilinearForm (and q odd)

list:=[];
for q in [3,5,7,9,11] do for t in [2,3,4] do for n in [1,2,3,4] do
f1:=GF(q^t);
f2:=GF(q);
#ps1:=HyperbolicQuadric(2*n-1,f1);
ps1:=EllipticQuadric(2*n-1,f1);
qf1:=SesquilinearForm(ps1);
qf2:=BilinearFormByFieldReduction(qf1,f2);
Add(list,PolarSpaceType(PolarSpace(qf2)));
od;od;od;
list; # contains only one type: the type of ps1

list:=[];
for q in [2,3,4,5,7,8,9] do for t in [2,3,4,5] do for n in [1,2,3,4] do
f1:=GF(q^t);
f2:=GF(q);
ps1:=SymplecticSpace(2*n-1,f1);
bil1:=SesquilinearForm(ps1);
bil2:=BilinearFormByFieldReduction(bil1,f2);
ps2:=PolarSpace(bil2);
Add(list,PolarSpaceType(ps2));
od;od;od;
list; 

list:=[];
for q in [3,5,7,9] do for t in [3,5] do for n in [1,2] do
f1:=GF(q^t);
f2:=GF(q);
#ps1:=HyperbolicQuadric(2*n-1,f1);
#ps1:=EllipticQuadric(2*n-1,f1);
ps1:=ParabolicQuadric(2*n,f1);
qf1:=QuadraticForm(ps1);
qf2:=QuadraticFormByFieldReduction(qf1,f2);
Add(list,PolarSpaceType(PolarSpace(qf2)));
od;od;od;
list; 

list:=[];
for q in [3,5,7,9] do for t in [3,5] do for n in [1,2] do
f1:=GF(q^t);
f2:=GF(q);
#ps1:=HyperbolicQuadric(2*n-1,f1);
#ps1:=EllipticQuadric(2*n-1,f1);
ps1:=ParabolicQuadric(2*n,f1);
qf1:=SesquilinearForm(ps1);
qf2:=BilinearFormByFieldReduction(qf1,f2);
Add(list,PolarSpaceType(PolarSpace(qf2)));
od;od;od;
list;

q := 2;
t := 5;
d1 := 4;
d2 := t*d1;
V2 := GF(q)^d2;
V1 := GF(q)^d1;
B := Basis(AsVectorSpace(GF(q),GF(q^t)));
basis := BasisVectors(B);
#shortvecs := List(V1);;
#longvecs := List(V2);
#List(longvecs,x->x=BlownUpVector(B,psi(x,q,t)));
ps1 := SymplecticSpace(3,GF(q^t));
ps1 := EllipticQuadric(3,GF(q^t));
shortvecs := List(Points(ps1),x->ElementToVectorSpace(x));;
shortspaces := List(Lines(ps1),x->ElementToVectorSpace(x));;

form1 := SesquilinearForm(ps1);
form1 := QuadraticForm(ps1);
form2 := BilinearFormByFieldReduction(form1,GF(q));
form2 := QuadraticFormByFieldReduction(form1,GF(q));
longspaces := List(shortvecs,x->BlownUpMat(B,[x]));;
longspaces2 := List(shortspaces,x->BlownUpMat(B,x));;
List(longspaces,x->IsTotallyIsotropicSubspace(form2,x));

List(longspaces2,x->IsTotallyIsotropicSubspace(form2,x));

longvects := List(shortvecs,x->BlownUpVector(B,x));
longspaces := List(shortvecs,x->BlownUpMat(B,[x]));;

